package mine;

import java.awt.*;
import java.util.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;

public class ImageLoader {

    private static Class c;

    /*** setup ******************************************************/

    public static void setup( Object obj ){
	c = obj.getClass();
    }

    /*** Load ****************************************************/

    public static BufferedImage getImage( Object fs, boolean alpha ){
	Image img =  loadImage(fs);
	if( img != null ){
	    int width = img.getWidth(null);
	    int height = img.getHeight(null);
	    return getBuffer( img, width, height, alpha );
	}else{
	    return getBuffer( null, 0, 0, alpha );
	}
    }

    /*** Create White Image *************************************/

    public static BufferedImage createImage( int width, int height, boolean alpha ){
	if( alpha )
	    return new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );
	else
	    return new BufferedImage( width, height, BufferedImage.TYPE_INT_RGB );
    }

    /*** Create Buffer ***************************************/

    public static BufferedImage getBuffer( Image img, int width, int height, boolean alpha ){

	if( img == null || width <= 0  || height <= 0 ){
	    return new BufferedImage( 1, 1, BufferedImage.TYPE_INT_RGB );
	}
	BufferedImage bimg;
	if( alpha )
	    bimg = new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );
	else
	    bimg = new BufferedImage( width, height, BufferedImage.TYPE_INT_RGB );

	Graphics g = bimg.getGraphics();
	g.drawImage( img, 0, 0, width, height, null );
	return bimg;
    }

    /*** Load Image *************************************************/

    private static Image loadImage( Object fs ){
	Image img = null;
	if( fs instanceof File ){
	    img = loadImageF((File)fs);
	}else if( fs instanceof String ){
	    img = loadImageS((String)fs);
	}
	if( img != null )
	    waitImage(img);
	else
	    System.out.println("File not found. ( " + fs + " )");

	return img;
    }

    /*** Load From File *************************************/

    private static Image loadImageF( File f ){
	if( f.exists() )
	    return Toolkit.getDefaultToolkit().getImage( f.getPath() );
	else
	    return null;
    }

    /*** Load By String ***********************************/

    private static Image loadImageS( String s ){
	if( c != null ){
	    URL rs = c.getResource(s);
	    if( rs != null )
		return Toolkit.getDefaultToolkit().getImage(rs);
	}
	return loadImageF( new File(s) );
    }

    /*** Wait Loading ******************************************/

    private static void waitImage( Image img ){
	try{
	    Toolkit tk = Toolkit.getDefaultToolkit();
	    for( int i=0; i<100; i++ ){
		if( tk.prepareImage(img, -1, -1, null) ) return;
		Thread.currentThread().sleep(100);
	    }
	}catch( Exception e ){
	    e.printStackTrace();
	}
    }

}
