package mine;

import java.awt.*;
import java.awt.image.*;
import java.util.Vector;

public class UnitMap {

    public static final int FALSE = 65535;

    protected int map[][][][];      // MapData

    protected int PageN;            // Page Number
    protected int mapX, mapY;       // Display Position
    protected int mapW, mapH;       // Map Size
    protected int unitW, unitH;     // Unit Size

    protected int unitX[], unitY[]; // Unit Position
    protected int numW[], numH[];   // Unit Number
    protected int spData[];         // Sprite Data

    protected Image unit[];         // Unit Image
    protected boolean Repaint[][];  // Repaint Flag
    protected boolean Visible[];    // Visible Flag
    protected int pxs, pys;         // Last Display Position

    /*** Constructer *************************************/

    public UnitMap(){}

    /*** Judge ***********************************************/

    protected boolean trueP( int p ){
	if( p < 0 )				return false;
	if( p >= PageN )	return false;
	return true;
    }
    protected boolean trueT( int t ){
	if( t < 0 )				return false;
	if( t >= 4 )			return false;
	return true;
    }
    protected boolean trueX( int x ){
	if( x < 0 )				return false;
	if( x >= mapW )		return false;
	return true;
    }
    protected boolean trueY( int y ){
	if( y < 0 )				return false;
	if( y >= mapH )		return false;
	return true;
    }
    protected boolean trueD( int p, int d ){
	if( d < 0 )										return false;
	if( d >= numW[p] * numH[p] )	return false;
	if( d == spData[p] )					return false;
	return true;
    }
    protected boolean trueA( int p, int t, int x, int y ){
	if( !trueP(p) ) return false;
	if( !trueT(t) ) return false;
	if( !trueX(x) ) return false;
	if( !trueY(y) ) return false;
	return true;
    }

    /*** Get and Set Data *****************************************/

    public int getPageNum(){
	return PageN;
    }
    public Point getLocation(){
	return new Point( mapX, mapY );
    }
    public Dimension getSize(){
	return new Dimension( mapW, mapH );
    }
    public Dimension getUnitSize(){
	return new Dimension( unitW, unitH );
    }
    public Point getUnitLocation( int p ){
	return new Point( unitX[p], unitY[p] );
    }
    public Dimension getUnitNum( int p ){
	return new Dimension( numW[p], numH[p] );
    }
    public void setLocation( int x, int y ){
	mapX = x;
	mapY = y;
    }
    public void setUnitSize( int w, int h ){
	unitW = w;
	unitH = h;
    }

    /*** Move Data *****************************************/

    public void keep( int p ){
	int[][] tmp = map[p][3];
	map[p][3] = map[p][2];
	map[p][2] = map[p][1];
	map[p][1] = tmp;
	this.copy( p, 0, p, 1 );
    }

    public void redo( int p ){
	int[][] tmp = map[p][0];
	map[p][0] = map[p][1];
	map[p][1] = map[p][2];
	map[p][2] = map[p][3];
	map[p][3] = tmp;
	repaint();
    }

    /*** Repaint *******************************************************/

    public void repaint(){
	repaint(true);
    }
    protected void repaint( boolean flag ){
	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		Repaint[x][y] = flag;
	    }
	}
    }

    /*** Range of Repaint *********************************************/

    public int[] getPaintBox( int p, int t, int px, int py, int mx, int my ){
	int sx = px + mx;
	int sy = py + my;
	int fx = px;
	int fy = py;
	for( int x=px;  x < px+mx; x++ ){
	    for( int y=py; y < py+my; y++ ){
		if( map[p][t][x][y] != 0 ){
		    if( x < sx ) sx=x;
		    if( y < sy ) sy=y;
		    if( x > fx ) fx=x;
		    if( y > fy ) fy=y;
		}
	    }
	}
	int[] box = new int[4];
	box[0] = sx - px;
	box[1] = sy - py;
	box[2] = fx - sx +1;
	box[3] = fy - sy +1;
	return box;
    }

    public int[] getPaintBox( int px, int py, int mx, int my ){
	int sx = px + mx;
	int sy = py + my;
	int fx = px;
	int fy = py;
	for( int x=px;  x < px+mx; x++ ){
	    for( int y=py; y < py+my; y++ ){
		if( Repaint[x][y] ){
		    if( x < sx ) sx=x;
		    if( y < sy ) sy=y;
		    if( x > fx ) fx=x;
		    if( y > fy ) fy=y;
		}
	    }
	}
	int[] box = new int[4];
	box[0] = sx - px;
	box[1] = sy - py;
	box[2] = fx - sx +1;
	box[3] = fy - sy +1;
	return box;
    }



    /*** VC **********************************************************************/

    public void C( int D01, int D02, int D03, int D04, int D05, int D06, int D07 ){

	PageN =	D01;
	mapX =	D02;
	mapY =	D03;
	mapW =	D04;
	mapH =	D05;
	unitW =	D06;
	unitH =	D07;

	map = new int[ PageN ][ 4 ][ mapW ][ mapH ];
	Repaint = new boolean[ mapW ][ mapH ];
	repaint();

	unitX =		new int[ PageN ];
	unitY =		new int[ PageN ];
	numW =		new int[ PageN ];
	numH =		new int[ PageN ];
	spData =	new int[ PageN ];
	unit =		new Image[ PageN ];
	Visible =	new boolean[ PageN ];
    }

    /*** VP **********************************************************************/

    public void P( int p, int D01, int D02, int D03, int D04, int D05, Image img ){

	unitX[p]	= D01;
	unitY[p]	= D02;
	numW[p]		= D03;
	numH[p]		= D04;
	spData[p]	= D05;
	unit[p]		= img;
    }
		

    /*** VV **********************************************************************/

    public void V( int p, boolean flag ){
	if( !trueP(p) ) return;
	if( Visible[p] != flag ){
	    Visible[p] = flag;
	    repaint();
	}
    }


    /*** VS **********************************************************************/

    protected void VS( int p, int t, int x, int y, int data ){
	if( map[p][t][x][y] == data ) return;
	map[p][t][x][y] = data;
	if( t == 0 ) Repaint[x][y] = true;
    }

    public void S( int p, int t, int x, int y, int data ){
	if( !trueA(p,t,x,y) ) return;
	VS( p,t,x,y,data );
    }

    /*** VG **********************************************************************/

    public int G( int p, int t, int x, int y ){
	if( !trueA(p,t,x,y) ) return FALSE;
	return map[p][t][x][y];
    }

    /*** VR **********************************************************************/

    public void R( int p, int t, int a[] ){

	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	int max = Math.min( a.length, mapW * mapH );

	for( int d=0;  d < max;  d++ ){
	    int x = d % mapW;
	    int y = d / mapW;
	    VS( p,t,x,y,a[d] );
	}
    }

    public void R( int p, int t, int a[][] ){

	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	int maxX = Math.min( mapW, a.length );
	int maxY = Math.min( mapH, a[0].length );
		
	for( int x=0; x < maxX; x++ ){
	    for( int y=0; y < maxY; y++ ){
		VS( p, t, x, y, a[x][y] );
	    }
	}
    }

    /*** VW *********************************************************************/

    public int[][] W( int p, int t ){
			
	if( !trueP(p) ) return null;
	if( !trueT(t) ) return null;
	if( mapW <= 0 ) return null;
	if( mapH <= 0 ) return null;

	int var[][] = new int[mapW][mapH];

	for( int x=0; x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		var[x][y] = map[p][t][x][y];
	    }
	}
	return var;
    }
		
    /*** VEP *********************************************************************/

    public synchronized void EP( int px, int py, int x, int y, Graphics g ){

	if( !trueX(x) ) return;
	if( !trueY(y) ) return;

	for( int p=0;  p < PageN; p++ ){
	    if( !Visible[p] ) continue;
	    int d = map[p][0][x][y];
	    if( !trueD(p,d) ) continue;
	    VE( p, x-px, y-py, d, g );
	}
	Repaint[x][y] = false;
    }

    /*** VE **********************************************************************/

    private void VE( int p, int x, int y, int d, Graphics g ){
	int axs = mapX + x * unitW;
	int ays = mapY + y * unitH;
	int axf = axs + unitW;
	int ayf = ays + unitH;
	int bxs = unitX[p] + (d % numW[p] ) * unitW;
	int bys = unitY[p] + (d / numW[p] ) * unitH;
	int bxf = bxs + unitW;
	int byf = bys + unitH;
	g.drawImage( unit[p], axs, ays, axf, ayf, bxs, bys, bxf, byf, null );
    }

    public synchronized void E( int px, int py, int mx, int my, boolean flag, Graphics g ){

	if( !flag || (px != pxs) || (py != pys) ) repaint();
	pxs = px;  pys = py;

	for( int x=0;  x < mx; x++ ){
	    for( int y=0;  y < my; y++ ){
		int xs = x + px;
		int ys = y + py;
		for( int p=0;  p < PageN; p++ ){
		    if( !Visible[p] ) continue;
		    if( !Repaint[xs][ys] ) continue;
		    int d = G( p,0,xs,ys );
		    if( !trueD(p,d) ) continue;
		    VE( p,x,y,d,g );
		}
		Repaint[xs][ys] = false;
	    }
	}
    }

    /*** VF *****************************************************/

    public void F( Graphics g ){
	E( 0, 0, mapW, mapH, false, g );
    }

    /*** VH **********************************************************************/

    public void H( int p, int x, int y, int max ){
  
  	if( !trueP(p) ) return;

	Vector dim = new Vector();
	clear( p, 3, FALSE );
 
	HS( p, x, y, 0, dim );
	while( !dim.isEmpty() ){
	    Point d = (Point)dim.firstElement();
	    dim.remove(0);
	    int h = map[p][3][d.x][d.y];
	    if( h >= max ) continue;
	    HS( p, d.x-1, d.y, h+1, dim );
	    HS( p, d.x+1, d.y, h+1, dim );
	    HS( p, d.x, d.y-1, h+1, dim );
	    HS( p, d.x, d.y+1, h+1, dim );
	}
    }
    private void HS( int p, int x, int y, int h, Vector dim ){
  	if( !trueX(x) ) return;
  	if( !trueY(y) ) return;
	if( map[p][2][x][y] != 0 )  return;
	if( map[p][3][x][y] <= h )  return;
	map[p][3][x][y] = h;
	dim.add( new Point( x,y ) );
    }

    /*** VJ ************************************************************/

    public void J( int p, int x, int y, int max ){
  
  	if( !trueP(p) ) return;

	Vector dim = new Vector();
	clear( p, 3, FALSE );

	JS( p, x, y, 0, max, dim );

	while( !dim.isEmpty() ){
	    Point d = (Point)dim.firstElement();
	    dim.remove(0);
	    int h = map[p][3][d.x][d.y];
	    JS( p, d.x-1, d.y, h, max, dim );
	    JS( p, d.x+1, d.y, h, max, dim );
	    JS( p, d.x, d.y-1, h, max, dim );
	    JS( p, d.x, d.y+1, h, max, dim );
	}
    }
    private void JS( int p, int x, int y, int h, int max, Vector dim ){
  	if( !trueX(x) ) return;
  	if( !trueY(y) ) return;
  	int n = h + map[p][2][x][y];             // current step
	if( n > max )  return;
	if( map[p][3][x][y] <= n )  return;
	map[p][3][x][y] = n;
	if( n == max ) return;
	dim.add( new Point( x,y ) );
    }

    /*** paint *********************************************************/

    public void paint( int p, int t, int x, int y, int cd ){
  	
  	if( !trueA(p,t,x,y) ) return;

	Vector dim = new Vector();
	int sd = map[p][t][x][y];
	if( cd == sd ) return;
	paint( p, t, x, y, sd, cd, dim );
	while( dim.size() != 0 ){
	    Point d = (Point)dim.firstElement();
	    dim.remove(0);
	    paint( p, t, d.x-1, d.y, sd, cd, dim );
	    paint( p, t, d.x+1, d.y, sd, cd, dim );
	    paint( p, t, d.x, d.y-1, sd, cd, dim );
	    paint( p, t, d.x, d.y+1, sd, cd, dim );
	}
    }
    private void paint( int p, int t, int x, int y, int sd, int cd, Vector dim ){
  	if( !trueX(x) ) return;
  	if( !trueY(y) ) return;
	if( map[p][t][x][y] != sd )  return;
	VS( p,t,x,y,cd );
	dim.add( new Point( x, y ) );
    }

    /*** fillRect **********************************************************************/

    public void fillRect( int p, int t, int px, int py, int mx, int my, int d ){
	
	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	int sx = Math.max( 0, px );
	int sy = Math.max( 0, py );
	int lx = Math.min( px+mx, mapW );
	int ly = Math.min( py+my, mapH );

	for( int x=sx; x<lx; x++ ){
	    for( int y=sy; y<ly; y++ ){
		VS( p,t,x,y,d );
	    }
	}
    }

    /*** drawRect ************************************************************************/

    public void drawRect( int p, int t, int px, int py, int mx, int my, int d ){
	drawBox( p, t, px, py, px + mx -1, py + my -1, d );
    }

    /*** boxFill *************************************************************************/

    public void fillBox( int p, int t, int ax, int ay, int bx, int by, int d ){
	int px = Math.min( ax, bx );
	int py = Math.min( ay, by );
	int mx = Math.abs( ax - bx );
	int my = Math.abs( ay - by );
	fillRect( p, t, px, py, mx, my, d );
    }

    /*** boxLine *************************************************************************/

    public void drawBox( int p, int t, int ax, int ay, int bx, int by, int d ){

	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	int sx = Math.min( ax, bx );
	int sy = Math.min( ay, by );
	int lx = Math.max( ax, bx );
	int ly = Math.max( ay, by );

	for( int x = sx;  x <= lx;  x++ ){
	    if( !trueX(x) ) continue;
	    if( trueY(sy) ) VS( p,t,x,sy,d );
	    if( trueY(ly) ) VS( p,t,x,ly,d );
	}
	for( int y = sy;  y <= ly;  y++ ){
	    if( !trueY(y) ) continue;
	    if( trueX(sx) ) VS( p,t,sx,y,d );
	    if( trueX(lx) ) VS( p,t,lx,y,d );
	}
    }


    /*** fillDia ************************************************************/

    public void fillDia( int p, int t, int x, int y, int r, int d ){

	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	for( int i=0; i<=r; i++ ){
	    for( int j=i-r; j<=r-i; j++ ){
		if( !trueX( x+j ) ) continue;
		if( trueY( y-i ) ) VS( p,t,x+j,y-i,d );
		if( trueY( y+i ) ) VS( p,t,x+j,y+i,d );
	    }
	}
    }

    /*** drawDia ***********************************************************/

    public void drawDia( int p, int t, int x, int y, int r, int d ){
	
	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	for( int i=0; i<=r; i++ ){
	    int j = r-i;
	    if( trueX( x+i ) ){
		if( trueY( y-j ) ) VS( p,t,x+i,y-j,d );
		if( trueY( y+j ) ) VS( p,t,x+i,y+j,d );
	    }
	    if( trueX( x-i ) ){
		if( trueY( y-j ) ) VS( p,t,x-i,y-j,d );
		if( trueY( y+j ) ) VS( p,t,x-i,y+j,d );
	    }
	}
    }

    /*** copy **********************************************************************/

    public void copy( int ap, int at, int bp, int bt ){

	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		VS( bp, bt, x, y, map[ap][at][x][y] );
	    }
	}
    }

    public void copy( int ap, int at, int ax, int ay, int mx, int my, int bp, int bt, int bx, int by ){

	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0; x < mx; x++ ){
	    if( !trueX(ax+x) ) continue;
	    if( !trueX(bx+x) ) continue;
	    for( int y=0; y < my; y++ ){
		if( !trueY(ay+y) ) continue;
		if( !trueY(by+y) ) continue;
		VS( bp, bt, bx+x, by+y, map[ap][at][ax+x][ay+y] );
	    }
	}
    }

    /*** clear **********************************************************************/

    public void clear( int p, int t, int d ){

	if( !trueP(p) ) return;
	if( !trueT(t) ) return;

	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		VS( p,t,x,y,d );
	    }
	}
    }

    /*** change **********************************************************************/

    public void change( int ap, int at, int ad, int bp, int bt, int bd ){

	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		if( map[ap][at][x][y] != ad ) continue;
		VS( bp,bt,x,y,bd );
	    }
	}
    }

    public void change( int ap, int at, int ad, int bp, int bt, int bd, int be ){

	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		int d = ( map[ap][at][x][y] == ad )? bd : be;
		VS( bp,bt,x,y,d );
	    }
	}
    }

    public void change( int ap, int at, int bp, int bt, int[] d ){

	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0;  x < mapW; x++ ){
	    for( int y=0; y < mapH; y++ ){
		int a = map[ap][at][x][y];
		if( a >= d.length ) continue;
		VS( bp,bt,x,y,d[a] );
	    }
	}
    }

    /*** affine *********************************************/

    public void affine( int ap, int at, int ax, int ay, int mx, int my, 
			int bp, int bt, int bx, int by, int[] af ){
	if( af == null ) return;
	if( af.length < 6 ) return;
	if( !trueP(ap) ) return;
	if( !trueT(at) ) return;
	if( !trueP(bp) ) return;
	if( !trueT(bt) ) return;

	for( int x=0; x < mx; x++ ){
	    if( !trueX(ax+x) ) continue;
	    for( int y=0; y < my; y++ ){
		if( !trueY(ay+y) ) continue;
		int xs = x*af[0] + y*af[1] + af[2];
		int ys = x*af[3] + y*af[4] + af[5];
		if( !trueX(bx+xs) ) continue;
		if( !trueY(by+ys) ) continue;
		VS( bp, bt, xs, ys, map[ap][at][ax+x][ay+y] );
	    }
	}
    }
}
